/** **********************************************************************
 * ConsoleApplet.java
 * A basic question-and-answer dialog GUI
 *
 * $Id$
 * (C) 1998 E.J. Friedman-Hill and the Sandia Corporation
 ********************************************************************** */

package jess;

import java.awt.*;
import java.applet.*;
import java.util.*;
import java.io.*;

//////////////////////////////////////////////////////////////////////
//
//    An Applet which uses ConsoleDisplay. It could
//    serve as the basis for any number of 'interview' style Expert System
//    GUIs.
//
//    Applet Parameters:
//    INPUT: if present, the applet will find the named document relative
//    to the applet's document base and interpret the file in batch mode,
//    then fall into a parse loop when the file completes.
//    COMPACT: If present, only the intrinsic functions are loaded. If
//    absent, all available userfunctions will be loaded.
// //////////////////////////////////////////////////////////////////////

public class ConsoleApplet extends Applet implements Runnable
{
  // The actual ReteDisplay
  private ConsoleDisplay m_cd;
  // The inference engine
  private Rete m_rete;
  // A parser
  private Jesp m_jesp;
  // Thread in which the parse loop runs
  private Thread m_thread;
  
  public void init()
  {
    setLayout(new BorderLayout());
    m_cd = new ConsoleDisplay(this);
    add("Center", m_cd.panel());
    add("South", new Label());
    m_rete = new Rete(m_cd);
    m_jesp = new Jesp(m_cd.stdin(), m_rete);
    
    // Process Applet Parameters
    String appParam = getParameter("INPUT");
    if (appParam != null)
      setCommand("(batch \"" + appParam + "\")");

    appParam = getParameter("COMPACT");
    if (appParam == null)
      {
        // Load in optional packages, but don't fail if any are missing.
        String [] pkgs = { "jess.StringFunctions",
                               "jess.PredFunctions",
                               "jess.MultiFunctions", 
                               "jess.MiscFunctions",
                               "jess.MathFunctions",
                               "jess.BagFunctions",
                               "jess.reflect.ReflectFunctions",
                               "jess.view.ViewFunctions" };
        
        for (int i=0; i< pkgs.length; i++)
          {
            try
              {
                m_rete.addUserpackage((Userpackage)
                                      Class.forName(pkgs[i]).newInstance());
              }
            catch (Throwable t) { /* Optional package not present */ }
          }
      }
        
    try
      {
        m_rete.executeCommand("(printout t \"Jess, the Java Expert System " +
                              "Shell\" crlf)");
        m_rete.executeCommand("(printout t \"Copyright (C) 1998 E.J. " +
                              "Friedman-Hill and  the Sandia Corporation\" crlf)");
        
        m_rete.executeCommand("(printout t (jess-version-string) crlf crlf)");
        m_rete.executeCommand("(printout t \"Jess output appears in this " +
                              "window.\" crlf)");
        m_rete.executeCommand("(printout t \"Type your input in the " +
                              "one-line input box below.\" crlf)");

      }
    catch (ReteException re)
      {
        m_cd.stdout().println("ERROR DURING SETUP: " + re.toString());
      }
  }

  private String m_cmd;

  public synchronized void setCommand(String cmd)
  {
    m_cmd = cmd;
  }

  public synchronized void run()
  {
    try
      {
        if (m_cmd != null)
          {
            m_rete.executeCommand(m_cmd);
            m_cmd = null;
          }
        
        // Loop and run commands.
        do 
          {
            try
              {
                m_jesp.parse(true);
              }
            catch (ReteException re) 
              {
                re.printStackTrace(m_rete.errStream());
              }
            catch (ArrayIndexOutOfBoundsException ai) 
              {
                m_rete.errStream().println("Wrong # of args in funcall?");
                ai.printStackTrace(m_rete.errStream());
              }
            catch (NullPointerException npe) 
              {
                m_rete.errStream().println("Wrong # of args in funcall?");
                npe.printStackTrace(m_rete.errStream());
              }
            catch (Exception e) 
              {
                m_rete.errStream().println("Unexpected exception:");
                e.printStackTrace(m_rete.errStream());
              }
          }
        while (true);
      }         
    catch (Throwable t) { m_rete.errStream().println(t); }
    m_thread = null;
  }

  public void start()
  {
    if (m_thread == null)
      {
        m_thread = new Thread(this);
        m_thread.start();
      }
  }

  public void stop()
  {
    if (m_thread != null)
      {
        m_thread.stop();
        m_thread = null;
      }
  }

}

