/** **********************************************************************
 * An execution context for Funcalls. To be used as a base class for Defrule,
 * Deffunction, etc.
 *
 * $Id$
 * (C) 1998 E.J. Friedman-Hill and the Sandia Corporation
 ********************************************************************** */

package jess;

import java.util.*;

public class Context 
{
  private Vector m_actions;
  private Hashtable m_bindings;
  private Context m_parent;  
  private boolean m_return;
  private Value m_retval;
  Rete m_engine;
  
  Context(Context c) 
  {
    m_engine = c.m_engine;
    m_parent = c;
  }

  Context(Rete engine)
  {
    m_engine = engine;
    m_parent = null;
  }
  
  final boolean returning() 
  {
    return m_return;
  }

  final Value setReturnValue(Value val) 
  {
    m_return = true;
    m_retval = val;
    return val;
  }

  final Value getReturnValue() 
  {
    return m_retval;
  }

  final void clearReturnValue() 
  {
    m_return = false;
    m_retval = null;
  }

  final int nActions()
  {
    if (m_actions == null)
      return 0;
    else
      return m_actions.size();
  }

  final Vector getActions()
  {
    if (m_actions == null)
      m_actions = new Vector();

    return m_actions;
  }

  final int nBindings()
  {
    if (m_bindings == null)
      return 0;
    else
      return m_bindings.size();
  }

  final Hashtable getBindings()
  {
    if (m_bindings == null)
      m_bindings = new Hashtable();

    return m_bindings;
  }

  public final Rete engine() 
  {
    return m_engine; 
  }

  Context push() 
  {    
    return new Context(this);
  }
  
  Context pop() 
  {
    if (m_parent != null)
      {
        m_parent.m_return = m_return;
        m_parent.m_retval = m_retval;        
        return m_parent;
      }
    else
      return this;
  }

  /**
    Find a variable table entry by 'chaining' upwards
    */
  
  Binding findBinding(String key) 
  {
    Context c = this;
    while (c != null)
      {
        Binding b = (Binding) c.getBindings().get(key);
        if (b != null)
          return b;
        else
          c = c.m_parent;
      }
    return null;
  }

  /**
    Add a Funcall to this context
    */

  final void addAction(ValueVector funcall) 
  {
    getActions().addElement(funcall);
  }


  /**
    Make note of a variable binding during parsing,
    so it can be used at runtime; the fact and slot indexes are for the use
    of the subclasses. Defrules use factIndex and slotIndex to indicate where in
    a fact to get data from; Deffunctions just use factIndex to indicate which
    element of the argument vector to pull their information from.
    */

  final Binding addBinding(String name, int factIndex, int slotIndex,
                           int subIndex) 
  {
    Binding b = findBinding(name);
    if (b == null)
      {
        b = new Binding(name, factIndex, slotIndex, subIndex);
        m_bindings.put(name, b);
      }
    return b;
  }

  /**
    Set a (possibly new!) variable to some type and value
    */

  final Binding setVariable(String name, Value value) 
  {
    Binding b = findBinding(name);
    if (b == null)
      b = addBinding(name, RU.LOCAL, RU.LOCAL, -1);
    b.m_val = value;

    return b;
  }

  /*
    Return binding from highest Context on chain
    */

  Binding findGlobalBinding(String key)
  {
    Context c = this;
    while (c.m_parent != null)
      c = c.m_parent;

    return (Binding) c.getBindings().get(key);
  }

  /*
    Add binding to highest Context on chain
    */

  final public Binding addGlobalBinding(String name, Value value)
  {
    Binding b;
    if ((b = findGlobalBinding(name)) != null) 
      {
        b.m_val = value;
      }
    else
      {
        b = new Binding(name, value);
        Context c = this;
        while (c.m_parent != null)
          c = c.m_parent;
        c.getBindings().put(name, b);
      }
    return b;
  }
  
  /**
    Recursive function which expands an action ValueVector by doing
    variable substitutions.
    @returns A copy of the action, expanded.
    */

  Funcall expandAction(Funcall original) throws ReteException 
  {
    // Make a local copy 
    Funcall action = (Funcall) original.clone();
    String functor = action.get(0).stringValue();            

    boolean isForeach = (functor.equals("foreach"));
    boolean bindOrForeach = 
      (functor.equals("bind") || isForeach);

    Userfunction uf = action.m_funcall;
    if (uf == null)
      {
        if ((uf = original.m_funcall =
             m_engine.findUserfunction(functor)) == null)
          throw new ReteException("Context::expandAction",
                                  "Unimplemented function",
                                  functor);
      }
    
    int firstUnexpandedArg;

    if (uf instanceof ControlStructure)
      {
        firstUnexpandedArg = ((ControlStructure) uf).skipFromHere();
      }
    else
      firstUnexpandedArg = action.size();

    // do variable replacements, using current values;
    // note that these can change during a rule execution!
    for (int j=1; j<firstUnexpandedArg; j++) 
      {

        // -*- sigh -*-
        // For very few forms, a variable is treated as an lvalue and
        // therefore must not be substituted here. For now, only
        // bind and foreach are treated this way: Just leave the first arg
        // alone. Note that the actual bind is still executed in Funcall.
      
        Value current = action.get(j);      
        if (bindOrForeach && j == 1)
          {
            String name = current.stringValue();
            Binding b = findBinding(name);
            if (b == null) 
              // This had better be an lvalue...
              b = setVariable(name, null);
            continue;
          }

        action.set(expandValue(current), j);
      }
    return action;
  }
  
  public Value expandValue(Value current) throws ReteException 
  {

    switch (current.type()) 
      {
      case RU.VARIABLE:
      case RU.MULTIVARIABLE:
        {
          Binding b = findBinding(current.stringValue());
          if (b == null) 
            {
              // This had better be an lvalue...
              b = setVariable(current.stringValue(), null);
            }
          else 
            {
              // plug in current values. 
              if (b.m_val != null)
                return b.m_val;
              else
                return Funcall.NIL();
            }
          break;
        } 
      case RU.FUNCALL:
        {
          Funcall subaction = current.funcallValue();
          return new Value(expandAction(subaction), RU.FUNCALL);
        }
      case RU.ORDERED_FACT:
      case RU.UNORDERED_FACT:
        {
          ValueVector fact = current.factValue();
          return new Value(expandFact(fact), current.type());
        }
      case RU.LIST:
        {
          ValueVector list = current.listValue();
          list = expandList(list);
          list = flattenList(list);
          return new Value(list, current.type());
        }
      }
    return current;
  }


  /**
    Do varsubs in facts
    */
  
  public ValueVector expandFact(ValueVector original) throws ReteException
  {
    ValueVector fact = (ValueVector) original.clone();

    for (int i=RU.FIRST_SLOT; i<fact.size(); i++) 
      {
        switch (fact.get(i).type()) 
          {
          case RU.VARIABLE:
          case RU.MULTIVARIABLE:
            {
              Binding b = findBinding(fact.get(i).stringValue());
              fact.set(b.m_val, i);
              break;
            }
          case RU.FUNCALL:
            {
              Funcall fc = expandAction(fact.get(i).funcallValue());
              Value v = Funcall.execute(fc, this, null, null);
              fact.set(v, i);
              break;
            }
          case RU.LIST:
            {
              ValueVector list = fact.get(i).listValue();
              list = expandList(list);
              list = flattenList(list);
              fact.set(new Value(list, RU.LIST), i);
              break;
            }
          }
      }
    return fact;
  }

  /**
    Do varsubs in lists
    */

  ValueVector expandList(ValueVector original) throws ReteException 
  {

    ValueVector list = (ValueVector) original.clone();

    for (int i=0; i<list.size(); i++) 
      {
        switch (list.get(i).type()) 
          {
          case RU.VARIABLE:
          case RU.MULTIVARIABLE:
            {
              Binding b = findBinding(list.get(i).stringValue());
              list.set(b.m_val, i);
              break;
            }
          case RU.FUNCALL:
            {
              Funcall fc = expandAction(list.get(i).funcallValue());
              Value v = Funcall.execute(fc, this, null, null);
              list.set(v, i);
              break;
            }
          case RU.LIST:
            {
              ValueVector vv = expandList(list.get(i).listValue());
              list.set(new Value(vv, RU.LIST), i);
              break;
            }
          }
      }
    
    return list;
  }
  
  // Replace a nested set of lists with a flat list.

  static final ValueVector flattenList(ValueVector vv) throws ReteException 
  {
    ValueVector flat = new ValueVector();
    doFlattenList(flat, vv);
    return flat;
  }

  // Recursively smoosh nested lists into one flat list. 

  private static void doFlattenList(ValueVector acc, ValueVector vv)
       throws ReteException 
  {
    for (int i=0; i< vv.size(); i++) 
      {
        Value current = vv.get(i);
        if (current.type() != RU.LIST) 
          {        
            acc.add(current);
          }
        else
          doFlattenList(acc, current.listValue());
      }
  }
}
