/** **********************************************************************
 * Class used to represent Deffunctions.
 * A Deffunction has no ValueVector representation; it is always a
 * class object.
 *
 * $Id$
 * (C) 1998 E.J. Friedman-Hill and the Sandia Corporation
 ********************************************************************** */
package jess;

import java.util.*;

/**

@author E.J. Friedman-Hill (C)1996
*/

public class Deffunction implements Userfunction
{
  
  String m_name;
  String m_docstring;
  private int m_nargs;
  private Hashtable m_bindings = new Hashtable();
  private Vector m_actions = new Vector();

  /**
    Constructor.
    */

  public final String name() { return m_name; }
  public final String docstring() { return m_docstring; }

  Deffunction(String name)
  {
    m_name = name;
  }

  /**
    Add a formal argument to this deffunction.
    */

  void addArgument(String name, int type) 
  {
    m_bindings.put(name, new Binding(name, m_nargs + 1, type, -1));
    ++m_nargs;
  }

  /**
    Add a simple value to this deffunction.
    */

  void addValue(Value val) 
  {
    m_actions.addElement(val);
  }

  /**
    Add an action to this deffunction
    */

  void addAction(Funcall fc) 
  {
    m_actions.addElement(fc);
  }

  /**
    Execute this deffunction. For each action (ValueVector form of a Funcall),
    do two things:
    1) Call ExpandAction to do variable substitution and
    subexpression expansion
    2) call Funcall.Execute on it.


    Vars is the ValueVector form of the funcall that we're executing in.
    */

  public Value call(ValueVector call, Context c) throws ReteException 
  {
    
    Value result = Funcall.NIL();

    if (call.size() < (m_nargs + 1))
      throw new ReteException("Deffunction::call",
                              "Too few arguments to Deffunction",
                              m_name);

    c = new Context(c.m_engine.globalContext());
    c.clearReturnValue();

    // set up the variable table
    for (Enumeration e = m_bindings.elements(); e.hasMoreElements();)
      {
        Binding b = (Binding) e.nextElement();
        switch (b.m_slotIndex)
          {
            // No default bindings for locals
          case RU.LOCAL:
            c.setVariable(b.m_name, Funcall.NIL());
            continue;

            // all others variables come from arguments
          case RU.VARIABLE:            
            c.setVariable(b.m_name, call.get(b.m_factIndex));
            break;
          case RU.MULTIVARIABLE:
            {
              ValueVector vv = new ValueVector();              
              for (int i=b.m_factIndex; i< call.size(); i++)
                vv.add(call.get(i));
              c.setVariable(b.m_name, new Value(vv, RU.LIST));
              break;
            }
          }
      }

    // OK, now run the function. For every action...
    try
      {
        int size = m_actions.size();
        for (int i=0; i<size; i++) 
          {
            // do all substitutions on this action
            if (m_actions.elementAt(i) instanceof ValueVector) 
              {
                Funcall original = (Funcall) m_actions.elementAt(i);
                Funcall action = c.expandAction(original);
                result = Funcall.execute(action, c);
              }
            else 
              {
                // it's a value
                Value v = (Value) m_actions.elementAt(i);
                result = c.expandValue(v);
              }
            
            if (c.returning()) 
              {
                result = c.getReturnValue();
                c.clearReturnValue();
                break;
              }
          }
      }
    catch (ReteException re)
      {
        re.addContext("deffunction " + m_name);
        throw re;
      }
    finally
      {
        c.pop();
      }
    return result;
  }
  
  /**
    Describe myself
    */

  public String toString() 
  {
    StringBuffer sb = new StringBuffer(100);
    sb.append("[Deffunction: ");
    sb.append(m_name);
    sb.append(" ");
    if (m_docstring != null)
      sb.append("\"" + m_docstring + "\"; ");
    sb.append("]");
    return sb.toString(); 
  }

}

