/** **********************************************************************
 * Class used to represent Defrules.
 *
 * $Id$
 * (C) 1998 E.J. Friedman-Hill and the Sandia Corporation
 ********************************************************************** */

package jess;

import java.util.*;
import java.io.*;

public class Defrule
{
  
  String m_name;
  Vector m_patts;
  String m_docstring;
  private int m_id;
  private int m_salience;
  private Value m_salienceVal;
  private boolean m_frozen = false;
  private Funcall[] m_localActions;
  private Vector m_nodes = new Vector();
  private Hashtable m_bindings = new Hashtable();
  private Vector m_actions = new Vector();
  private Rete m_engine;
  public final String name() { return m_name; }
  public final int salience() { return m_salience; }

  void salience(Value i) throws ReteException
  { m_salienceVal = i; evalSalience();}

  public final String docstring() { return m_docstring; }

  public int evalSalience() throws ReteException
  {
    Context gc = m_engine.globalContext().push();
    try
      {
        Value v = gc.expandValue(m_salienceVal);
        switch (v.type())
          {
          case RU.INTEGER:
          case RU.FLOAT:
            m_salience = v.intValue(); break;
          case RU.FUNCALL:
            {
              Value rv = Funcall.execute(v.funcallValue(), gc);
              m_salience = rv.intValue();
            }
          
          }
      }
    finally
      {
        gc.pop();    
      }
    return m_salience;
  }

  /**
    Constructor.
    */

  Defrule(String name, Rete engine) throws ReteException
  {
    m_engine = engine;
    m_name = name;
    m_id = engine.nextRuleId();
    m_patts = new Vector();
    m_salience = 0;
    m_salienceVal = new Value(0, RU.INTEGER);
  }

  /**
    The id is read-only
    */

  public int id() { return m_id; }

  /**
    Add a pattern to this Defrule
    */

  void addPattern(Pattern pattern) throws ReteException 
  {
    Hashtable substitutes = new Hashtable();
    m_patts.addElement(pattern);
    pattern.compact();


    // Look for variables in the fact, and create bindings for new ones
    for (int i=0; i< pattern.m_tests.length; i++)
      {
        if (pattern.m_tests[i] == null)
          continue;
        for (int j=0; j< pattern.m_tests[i].length; j++) 
          {
            Test1 test = pattern.m_tests[i][j];
            Value val = test.m_slotValue;
            boolean eq = (test.m_test == Test1.EQ);
            if ((val.type() & (RU.VARIABLE | RU.MULTIVARIABLE)) != 0)
              {
                String name = val.stringValue();
                if (m_bindings.get(name) == null)
                  {
                    if (pattern.negated() != 0 && !name.startsWith("_blank"))
                      // "Disguise" this variable name; include patt #
                      // so patterns are independent
                      {
                        String oldName = name;
                        name = "_" + m_patts.size() + "_" + name;
                        test.m_slotValue = new Value(name, val.type());
                        substitutes.put(oldName, name);
                      }

                    else if (!eq) 
                      {
                      throw new ReteException("Defrule::addPattern",
                                              "First use of variable negated",
                                              name);
                      }

                    m_bindings.put(name, new Binding(name, m_patts.size() - 1,
                                                     RU.FIRST_SLOT +i, test.m_subIdx)); 
                  }
              }
          }
      }

    // Now go back and rename any variables that need renaming
    for (int i=0; i< pattern.m_tests.length; i++)
      {
        if (pattern.m_tests[i] == null)
          continue;
        for (int j=0; j< pattern.m_tests[i].length; j++) 
          {
            Test1 test = pattern.m_tests[i][j];
            Value val = test.m_slotValue;
            if (val.type() == RU.FUNCALL)
              substFuncall(val.funcallValue(), substitutes);
          }
      }
  }

  private void substFuncall(Funcall f, Hashtable substitutes) throws ReteException
  {
    for (int i=1; i<f.size(); i++)
      {      
        Value current = f.get(i);
        if ((current.type() & (RU.VARIABLE | RU.MULTIVARIABLE)) != 0)
          {
            String s = (String) substitutes.get(current.stringValue());
            if (s != null)
              f.set(new Value(s, current.type()), i);
          }
        else if (current.type() == RU.FUNCALL)        
          substFuncall(current.funcallValue(), substitutes);
      }
  }

  /**
    Add an action to this deffunction
    */

  void addAction(Funcall fc) 
  {
    m_actions.addElement(fc);
  }

  void addBinding(String name, int factidx, int slotidx, int subidx) 
  {
    m_bindings.put(name, new Binding(name, factidx, slotidx, subidx));
  }

  Hashtable getBindings() { return m_bindings; }

  /**
    Tell this rule to set the actions up for faster execution
    */

  void freeze() 
  {
    if (m_frozen) return;

    m_frozen = true;

    m_localActions = new Funcall[m_actions.size()];
    for (int i=0; i<m_localActions.length; i++)
      m_localActions[i] = (Funcall) m_actions.elementAt(i);

    
  }
  

  void ready(Token fact_input, Context c) throws ReteException 
  {

    ValueVector fact;
    // set up the variable table    
    for (Enumeration e = m_bindings.elements(); e.hasMoreElements();) 
      {
        Binding b = (Binding) e.nextElement();
        if (b.m_slotIndex == RU.LOCAL)
          // no default binding for locals
          continue;

        // all others variables need info from a fact
        // if this is a not CE, skip it;
        fact = fact_input.fact(b.m_factIndex);
        try 
          {
            if (b.m_slotIndex == RU.PATTERN) 
              {
                b.m_val = fact.get(RU.ID);
              } 
            else 
              {
                if (b.m_subIndex == -1)
                  {
                    b.m_val = fact.get(b.m_slotIndex);
                  }
                
                else 
                  {
                    ValueVector vv = fact.get(b.m_slotIndex).listValue();
                    b.m_val = vv.get(b.m_subIndex);
                  }
            
              }
            c.setVariable(b.m_name, b.m_val);
          }
        catch (Throwable t) 
          {
            // bad binding. These can come from unused bindings in not CE's.
          }     
      }
    return;
  }
  

  /**
    Do the RHS of this rule.For each action (ValueVector form of a Funcall),
    do two things:
    1) Call ExpandAction to do variable substitution and
    subexpression expansion
    2) call Funcall.Execute on it.

    Fact_input is the Vector of ValueVector facts we were fired with.
    */

  void fire(Token fact_input) throws ReteException 
  {
    if (m_engine.watchRules())
      debugPrint(fact_input);
    
    Context c = new Context(m_engine.globalContext());
    c.clearReturnValue();

    // Pull needed values out of facts into bindings table
    ready(fact_input, c);

    try
      {
        // OK, now run the rule. For every action...
        int size = m_localActions.length;
        for (int i=0; i<size; i++) 
          {
            
            // do all substitutions on this action
            Funcall original = m_localActions[i];
            Funcall action = c.expandAction(original);
            Funcall.execute(action, c);
            
            if (c.returning()) 
              {
                c.clearReturnValue();
                c.pop();
                return;
              }
          }
      }
    catch (ReteException re)
      {
        re.addContext("defrule " + m_name);
        throw re;
      }
    finally
      {
        c.pop();
      }
    
  }
  

  void debugPrint(Token facts) throws ReteException 
  {
    PrintStream ps = m_engine.outStream();
    ps.print("FIRE " + toString());
    for (int i=0; i<facts.size(); i++) 
      {
        ValueVector f = facts.fact(i);
        if (f.get(RU.ID).factIDValue() != -1)
          ps.print(" f-" + f.get(RU.ID).factIDValue());
        if (i< facts.size() -1)
          ps.print(",");
      }
    ps.println();
  }
  
  /**
    Describe myself
    */

  public String toString() 
  {
    StringBuffer sb = new StringBuffer(100);
    sb.append("[Defrule: ");
    sb.append(m_name);
    sb.append(" ");
    if (m_docstring != null)
      sb.append("\"" + m_docstring + "\"; ");
    sb.append(m_patts.size());
    sb.append(" patterns; salience: ");
    sb.append(m_salience);
    sb.append("]");
    return sb.toString();
  }

  public String listNodes()
  {
    StringBuffer sb = new StringBuffer(100);
    for (int i=0; i< m_nodes.size(); i++)
      {
        sb.append(((Successor) m_nodes.elementAt(i)).m_node);
        sb.append("\n");
      }
    return sb.toString();
  }

  void addNode(Successor n)
  {
    m_nodes.addElement(n);
  }

  // ************************************************************
  // Completely remove this rule from the Rete network, including
  // removing any internal nodes that are only used by this rule.
  
  void remove(Vector roots)
  {
    Enumeration e = m_nodes.elements();
    while (e.hasMoreElements())
      {
        Successor s = (Successor) e.nextElement();
        if (--s.m_node.m_usecount <= 0)
          {
            roots.removeElement(s);
            Enumeration e2 = m_nodes.elements();
            while (e2.hasMoreElements())
              {
                Node n = ((Successor) e2.nextElement()).m_node;
                n.removeSuccessor(s);
              }
          }
      }
    // Straighten out the nodes we disturbed.
    e = m_nodes.elements();
    while (e.hasMoreElements())
      {
        Successor s = (Successor) e.nextElement();
        s.m_node.freeze();
      }
    m_nodes.removeAllElements();
  }

  /* **********************************************************************
   * following additions by Rajaram Ganeshan 12/29/97
   * to pretty print rules in JESS:
   * 1. public String ppRule()
   * 2. private String isBoundPattern( int factIndex )
   * 
   ********************************************************************** */

  public String ppRule() throws ReteException
  {
    String varname;
    StringBuffer sb = new StringBuffer(256);
    sb.append("(defrule ");
    sb.append(m_name);
    sb.append("\n");
    
    // Defrule docstrings are not required,
    // just encouraged...
    if ( m_docstring != null )
      {
        sb.append("   \"");
        sb.append(m_docstring);
        sb.append("\"\n");
      }  
    
    // check for salience declaration
    sb.append("(declare (salience ");
    if (m_salienceVal.type() == RU.FUNCALL)
      sb.append("(");
    sb.append(m_salienceVal);
    if (m_salienceVal.type() == RU.FUNCALL)
      sb.append(")");
    sb.append("))\n");
    
    // patterns for LHS
    for (int i=0; i<m_patts.size(); i++)
      {
        Pattern p = (Pattern) m_patts.elementAt(i);
        // Two cases are possible:
        // 1. pattern bound to a variable. These look like this:
        // ?name <- (pattern 1 2 3)
        //  2. pattern not bound to a val looks like:
        // (pattern 1 2 3)
        
        // ignore initial-fact if present
        if ( i == 0 && p.name().equals("initial-fact") )
          continue;
        
        // check if pattern is negated
        if (p.negated() != 0)
          {
            // negated patterns cannot be bound to a value
            sb.append("   (not ");
            sb.append(p.ppPattern(m_engine));
            sb.append(")\n");
          } 
        else 
          {
            sb.append("   ");
            // check if this pattern was bound to a val            
            if ( (varname = isBoundPattern( i )) != null )
              {
                sb.append("?");
                sb.append(varname);
                sb.append(" <- ");
              }
            
            if (p.unique())
              sb.append("(unique ");
            sb.append(p.ppPattern(m_engine));
            if (p.unique())
              sb.append(")");
            sb.append("\n");
          }
      }
    // "=>"
    sb.append("  =>\n");
    
    // actions for RHS
    for (int j=0; j < m_actions.size(); j++)
      {
        Funcall f = (Funcall) m_actions.elementAt(j);
        sb.append("   ");
        sb.append(f.ppFuncall(m_engine));
        sb.append("\n");
      }
    sb.append(")");
    return sb.toString();
  }
  
  private String isBoundPattern( int factIndex )
  {
    for (Enumeration e = m_bindings.elements(); e.hasMoreElements();)
      {
        Binding b = (Binding) e.nextElement();
        if (b.m_slotIndex == RU.PATTERN)
          if ( b.m_factIndex == factIndex )
                return b.m_name;
      }
    return null;
  }
  
}


