/** **********************************************************************
 *  Some small sample user-defined miscellaneous functions
 * 
 * To use one of these functions from Jess, simply register the
 * package class in your Java mainline:
 *
 * engine.AddUserpackage(new MiscFunctions());
 *
 * $Id$
 * (C) 1998 E.J. Friedman-Hill and the Sandia Corporation
 ********************************************************************** */

package jess;
import java.io.*;
import java.net.*;
import java.util.*;


public class MiscFunctions implements Userpackage 
{

  public void add(Rete engine) 
  {
    engine.addUserfunction(new SetStrategy());
    engine.addUserfunction(new JSocket());
    engine.addUserfunction(new JFormat());
    engine.addUserfunction(new Batch());
    engine.addUserfunction(new JSystem());
    engine.addUserfunction(new PPDefrule());
    engine.addUserfunction(new LoadPkg());
    engine.addUserfunction(new LoadFn());
    engine.addUserfunction(new Time());
    engine.addUserfunction(new Build());
    engine.addUserfunction(new Eval());
    engine.addUserfunction(new ListFunctions());
    engine.addUserfunction(new Agenda());
    // setgen added by Win Carus (9.19.97)
    engine.addUserfunction(new Setgen());
    engine.addUserfunction(new ResetGlobals(ResetGlobals.SET));
    engine.addUserfunction(new ResetGlobals(ResetGlobals.GET));
    engine.addUserfunction(new EvalSalience(EvalSalience.SET));
    engine.addUserfunction(new EvalSalience(EvalSalience.GET));
  }
}

class SetStrategy implements Userfunction 
{
  public String name() { return "set-strategy"; }
  
  public Value call(ValueVector vv, Context context) throws ReteException
  {
    Strategy s = null;
    try
      {
        s = (Strategy)
          Class.forName("jess." + vv.get(1).stringValue()).newInstance();
      }
    catch(Throwable t)
      {
        try
          {
            s = (Strategy) Class.forName(vv.get(1).stringValue()).newInstance();
          }
        catch (Throwable tt)
          {
            throw new ReteException("set-strategy", "Strategy class not found:",
                                    vv.get(1).stringValue());
          }
      }

    // return name of old strategy
    String rv = context.engine().setStrategy(s);
    return new Value(rv, RU.ATOM);
  }
}

class Batch implements Userfunction 
{
  public String name() { return "batch";}
  
  public Value call(ValueVector vv, Context context) throws ReteException
  {

    String filename = vv.get(1).stringValue();
    Value v = Funcall.FALSE();
    InputStream fis = null;
    try
      {
        if (context.engine().display().applet() == null)
          fis = new FileInputStream(filename);
        else
          {
            URL url
              = new URL(context.engine().display().applet().getDocumentBase(),
                        vv.get(1).stringValue());          
            fis = url.openStream(); 
          }         
        Jesp j = new Jesp(fis, context.engine());
        do
          {
            v = j.parse(false);
          }
        while (fis.available() > 0);
      }
    catch (IOException ex)
      {
        throw new ReteException("batch", "I/O Exception on file", "");
      }    
    finally
      {
        if (fis != null) try { fis.close(); } catch (IOException ioe) {}
      }
    return v;
  }
}

class Build implements Userfunction 
{
  public String name() { return "build";}
  
  public Value call(ValueVector vv, Context context) throws ReteException
  {
    String argument = vv.get(1).stringValue();
    return context.engine().executeCommand(argument);
  }
}

// These will diverge in the future!

class Eval implements Userfunction 
{
  public String name() { return "eval";}
  
  public Value call(ValueVector vv, Context context) throws ReteException
  {
    String argument = vv.get(1).stringValue();
    return context.engine().executeCommand(argument);
  }
}

class ListFunctions implements Userfunction 
{
  public String name() { return "list-function$";}
  
  public Value call(ValueVector vv, Context context) throws ReteException
  {
    ValueVector rv = new ValueVector(100);

    Enumeration e = context.engine().listFunctions();
    while (e.hasMoreElements() )
      rv.add(new Value(((Userfunction) e.nextElement()).name(), RU.ATOM));
    
    // Bubble-sort the names.
    int swaps;
    do 
      {
        swaps = 0;
        for (int i=0; i< rv.size() -1; i++)
          {
            Value v1 = rv.get(i);
            Value v2 = rv.get(i+1);
            if (v1.stringValue().compareTo(v2.stringValue()) > 0)
              {
                ++swaps;
                rv.set(v2, i);
                rv.set(v1, i +1);
              }
          }

      }
    while (swaps > 0);

    return new Value(rv, RU.LIST);
  }
}

class Agenda implements Userfunction 
{
  public String name() { return "agenda";}
  
  public Value call(ValueVector vv, Context context) throws ReteException
  {
    Rete r = context.engine();
    PrintStream outStream = r.outStream();
    int n=0;
    for (Enumeration e = r.listActivations(); e.hasMoreElements(); n++)
      outStream.println(e.nextElement());
    outStream.print("For a total of ");
    outStream.print(n);
    outStream.println(" activations.");
    return Funcall.NIL();
  }
}

class PPDefrule implements Userfunction 
{
  public String name() { return "ppdefrule";}
  
  public Value call(ValueVector vv, Context context) throws ReteException
  {
    String argument = vv.get(1).stringValue();
    Defrule dr = context.engine().findDefrule(argument);
    return new Value(dr.ppRule(), RU.STRING);
  }
}

class JSystem implements Userfunction 
{
  public String name() { return "system";}
  
  public Value call(ValueVector vv, Context context) throws ReteException
  {
    try
      {
        boolean async = false;
        int size = vv.size();
        if (vv.get(size - 1).stringValue().equals("&"))
            {
              async = true;
              --size;
            }

        String[] words = new String[size - 1];
        for (int i=1; i<size; i++)
          words[i-1] = vv.get(i).stringValue();
        Process p = Runtime.getRuntime().exec(words);
        try
          {
            if (!async)
              p.waitFor();
          }
        catch (InterruptedException ie)
          {
            return Funcall.FALSE();
          }
        return Funcall.TRUE();
      }
    catch (IOException ioe)
      {
        throw new ReteException("system", "I/O Exception", "");
      }
    catch (SecurityException se)
      {
        throw new ReteException("system", "Security Exception", "");
      }
  }
}

class LoadPkg implements Userfunction 
{
  public String name() { return "load-package";}
  
  public Value call(ValueVector vv, Context context) throws ReteException
  {
    String clazz = vv.get(1).stringValue();
    try
      {
        Userpackage up = (Userpackage) Class.forName(clazz).newInstance();
        context.engine().addUserpackage(up);
      }
    catch (ClassNotFoundException cnfe)
      {
        throw new ReteException("load-package", "Class not found", clazz);
      }
    catch (IllegalAccessException iae)
      {
        throw new ReteException("load-package", "Class is not accessible",
                                clazz);
      }
    catch (InstantiationException ie)
      {
        throw new ReteException("load-package", "Class cannot be instantiated",
                                clazz);
      }
    catch (ClassCastException cnfe)
      {
        throw new ReteException("load-package",
                                "Class must inherit from UserPackage", clazz);
      }
    return Funcall.TRUE();
  }
}

class LoadFn implements Userfunction 
{
  public String name() { return "load-function";}
  
  public Value call(ValueVector vv, Context context) throws ReteException
  {
    String clazz = vv.get(1).stringValue();
    try
      {
        Userfunction uf = (Userfunction) Class.forName(clazz).newInstance();
        context.engine().addUserfunction(uf);
      }
    catch (ClassNotFoundException cnfe)
      {
        throw new ReteException("load-function", "Class not found", clazz);
      }
    catch (IllegalAccessException iae)
      {
        throw new ReteException("load-function", "Class is not accessible",
                                clazz);
      }
    catch (InstantiationException ie)
      {
        throw new ReteException("load-function",
                                "Class cannot be instantiated",
                                clazz);
      }
    catch (ClassCastException cnfe)
      {
        throw new ReteException("load-function",
                                "Class must inherit from UserFunction", clazz);
      }
      
    return Funcall.TRUE();
  }
}

class Time implements Userfunction
{
  public String name() { return "time";}

  public Value call( ValueVector vv, Context context ) throws ReteException
  {
    return new Value( System.currentTimeMillis()/1000, RU.FLOAT );
  }
}

class JSocket implements Userfunction
{
  public String name( ) { return "socket";}

  public Value call( ValueVector vv, Context context ) throws ReteException
  {
    String host = vv.get(1).stringValue();
    int port = vv.get(2).intValue();
    String router = vv.get(3).stringValue();

    try
      {
        Socket sock = new Socket(host, port);
        Rete engine = context.engine();
        engine.addInputRouter(router, sock.getInputStream(), false);
        engine.addOutputRouter(router, sock.getOutputStream());
        return vv.get(3);
      }
    catch (IOException ioe)
      {
        throw new ReteException("socket::call", "I/O Exception", ioe.toString());
      }
  }
}

class Setgen implements Userfunction
{
  public String name( ) { return "setgen" ;}

  public Value call( ValueVector vv, Context context ) throws ReteException
  {
    int i = vv.get( 1 ).intValue( );
    RU.s_gensymIdx = (i > RU.s_gensymIdx) ? i : (RU.s_gensymIdx + 1);

    return Funcall.TRUE( );
  }
}

class ResetGlobals implements Userfunction
{
  public static final int SET=0, GET=1;
  private int m_cmd;

  public ResetGlobals(int cmd) { m_cmd = cmd; }
  public String name( )
  { return m_cmd == SET ? "set-reset-globals" : "get-reset-globals"; }
  
  public Value call( ValueVector vv, Context context ) throws ReteException
  {
    switch (m_cmd)
      {
      case SET:
        {
          Value v = null;
          Value v1 = vv.get(1); 
          if (v1.equals(Funcall.NIL()) || v1.equals(Funcall.FALSE()))
            {
              context.engine().setResetGlobals(false);
              v = Funcall.FALSE();
            }
          else
            {
              context.engine().setResetGlobals(true);
              v = Funcall.TRUE();
            }        
          return v;
        }

      default:
        return context.engine().getResetGlobals() ?
          Funcall.TRUE() : Funcall.FALSE();         
      }
  }
}

class EvalSalience implements Userfunction
{
  public static final int SET=0, GET=1;
  private static final String [] s_values = {"when-defined", "when-activated",
                                             "every-cycle"};
  private int m_cmd;

  public EvalSalience(int cmd) { m_cmd = cmd; }
  public String name( )
  { return m_cmd == SET ? "set-salience-evaluation"
      : "get-salience-evaluation"; }
  
  public Value call( ValueVector vv, Context context ) throws ReteException
  {
    switch (m_cmd)
      {
      case SET:
        {
          int old = context.engine().getEvalSalience();
          String s = vv.get(1).stringValue();
          for (int i=0; i<s_values.length; i++)
            if (s.equals(s_values[i]))
              {
                context.engine().setEvalSalience(i);
                return new Value(s_values[old], RU.ATOM);
              }          
          throw new ReteException("set-eval-salience", "Invalid value: " + s,
                                  "(valid values are when-defined, " +
                                  "when-activated, every-cycle)");
        }
      case GET:
      default: 
        {
          return new Value(s_values[context.engine().getEvalSalience()],
                           RU.ATOM);                    
        }
      }
  }
}

/*
  format is implemented in terms of the Format class from the Core Java CD
  which has the annoying limitation that it can only format one datum at
  a time! 
  */

class JFormat implements Userfunction
{
  public String name( ) { return "format" ;}

  public Value call( ValueVector vv, Context context ) throws ReteException
  {
    // individual formats in here
    StringBuffer fmtbuf = new StringBuffer(20);

    // The eventual return value
    StringBuffer outbuf = new StringBuffer(100);

    // The router where things go
    Value router = vv.get(1);

    // The format command string    
    String fmt = vv.get(2).stringValue();
    // an index into the above
    int ptr = 0;

    // index over arguments
    int validx = 3;

    char c = 0;
    while (ptr < fmt.length())
      {
        // Any leading non-format stuff
        while (ptr < fmt.length() && (c = fmt.charAt(ptr++)) != '%')
          outbuf.append(c);
        
        if (ptr >= fmt.length())
          break;

        // copy format to fmtbuf
        fmtbuf.setLength(0);
        fmtbuf.append(c);
        while (ptr < fmt.length() && (c = fmt.charAt(ptr++)) != '%'
                && !Character.isLetter(c))
          fmtbuf.append(c);
        if (c == 'n')
          {
            outbuf.append('\n');
            break;
          }
        else if (c == '%')
          {
            outbuf.append('%');
            break;
          }
        else
          fmtbuf.append(c);
        
        Format f = new Format(fmtbuf.toString());
        Value v;
        switch (f.fmt)
          {
          case 'd': case 'i': case 'o': case 'x': case 'X':
            v = vv.get(validx++);
            outbuf.append(f.form(v.intValue())); break;
          case 'f': case 'e': case 'E': case 'g': case 'G':
            v = vv.get(validx++);
            outbuf.append(f.form(v.floatValue())); break;
          case 'c':
            v = vv.get(validx++);
            switch (v.type())
              {
              case RU.ATOM: case RU.STRING:
                outbuf.append(f.form(v.stringValue().charAt(0))); break;
              default:
                outbuf.append(f.form((char) v.intValue())); break;
              }
            break;
          case 's':
            v = vv.get(validx++);                
            outbuf.append(f.form(v.stringValue())); break;
            
          default:
            throw new ReteException("_format::call",
                                    "Unknown format",
                                    fmtbuf.toString());
          }
      }

    String s = outbuf.toString();
    if (!router.equals(Funcall.NIL()))
      {
        String routerName = router.stringValue();
        OutputStream os = context.engine().getOutputRouter(routerName);
        if (os == null)
          throw new ReteException("_format::call", "Bad router", routerName);
        try
          {
            for (int i=0; i<s.length(); i++)
              os.write(s.charAt(i));
            os.flush();
          }
        catch (IOException ioe)
          {
            throw new ReteException("_format::call", "I/O Exception", ioe.toString());
          }
      }

    return new Value(s, RU.STRING);
  }
}
