/** **********************************************************************
 * Two-input nodes of the Rete network
 * Test that a slot from the left input and one
 * from the right input have the same
 * value and type.
 *
 * $Id$
 * (C) 1998 E.J. Friedman-Hill and the Sandia Corporation
 ********************************************************************** */

package jess;

import java.util.*;
import java.awt.*;

public class Node2 extends NodeTest
{
  
  /**
    The left and right token memories
    They are Vectors of Tokens
    */

  TokenTree m_left, m_right;
  
  private boolean m_unique;

  /**
    Constructor
    */

  Node2(Rete engine, boolean unique) 
  {
    super(engine);
    m_left = new TokenTree(51);
    m_right = new TokenTree(51);
    m_unique = unique;
  }

  void addTest(int test, int token_idx, int left_idx, int right_idx) 
  { 
    Test t;
    m_tests.addElement(t = new Test2Simple(test, token_idx,
                                           left_idx, right_idx));
  }

  void addTest(int test, int token_idx, int left_idx, int leftSub_idx,
               int right_idx, int rightSub_idx) {

    Test t;
    if (leftSub_idx == -1 && rightSub_idx == -1)
      m_tests.addElement(t = new Test2Simple(test, token_idx,
                                             left_idx, right_idx));
    else
      m_tests.addElement(t = new Test2Multi(test, token_idx, left_idx, 
                                            leftSub_idx, right_idx,
                                            rightSub_idx));
  }


  /**
    Do the business of this node.
    The 2-input nodes, on receiving a token, have to do several things,
    and their actions change based on whether it's an ADD or REMOVE,
    and whether it's the right or left input!
    <PRE>

    *********** For ADDs, left input:
    1) Look for this token in the left memory. If it's there, do nothing;
    If it's not, add it to the left memory.

    2) Perform all this node's tests on this token and each of the right-
    memory tokens. For any right token for which they succeed:

    3) a) append the right token to a copy of this token. b) do a
    CallNode on each of the successors using this new token.

    *********** For ADDs, right input:

    1) Look for this token in the right memory. If it's there, do nothing;
    If it's not, add it to the right memory.

    2) Perform all this node's tests on this token and each of the left-
    memory tokens. For any left token for which they succeed:

    3) a) append this  token to a copy of the left token. b) do a
    CallNode on each of the successors using this new token.

    *********** For REMOVEs, left input:
    
    1) Look for this token in the left memory. If it's there, remove it;
    else do nothing.

    2) Perform all this node's tests on this token and each of the right-
    memory tokens. For any right token for which they succeed:
    
    3) a) append the right token to a copy of this token. b) do a
    CallNode on each of the successors using this new token.

    *********** For REMOVEs, right input:
    
    1) Look for this token in the right memory. If it's there, remove it;
    else do nothing.

    2) Perform all this node's tests on this token and each of the left-
    memory tokens. For any left token for which they succeed:
    
    3) a) append this token to a copy of the left token. b) do a
    CallNode on each of the successors using this new token.

    </PRE>
    */
  
  boolean callNode(Token token, int callType) throws ReteException
  {
    // This is a special case. If we get a 'clear', we flush our memories,
    // then notify all our successors and return.
    
    if (token.m_tag == RU.CLEAR)
      {
        m_left = new TokenTree(51);
        m_right = new TokenTree(51);
        passAlong(token);
        return false;
      }

    // the four cases listed above are now implemented in order.

    try
      {
        switch (callType) 
          {
            
          case Node.LEFT:
            switch (token.m_tag) 
              {
                
              case RU.ADD:
              case RU.UPDATE:
                m_left.add(token);
                runTestsVaryRight(token, m_right);
                break;
                
              case RU.REMOVE:
                m_left.remove(token);
                runTestsVaryRight(token, m_right);
                break;
                
              default:
                throw new ReteException("Node2::callNode",
                                        "Bad tag in token",
                                        String.valueOf(token.m_tag));
              } // switch token.tag
            break; // case Node.LEFT;
            
            
          case Node.RIGHT:
            switch (token.m_tag) 
              {
                
              case RU.UPDATE:
              case RU.ADD:
                m_right.add(token);
                runTestsVaryLeft(token, m_left);
                break;
                
              case RU.REMOVE:
                m_right.remove(token);
                runTestsVaryLeft(token, m_left);
                break;
                
              default:
                throw new ReteException("Node2::callNode",
                                        "Bad tag in token",
                                        String.valueOf(token.m_tag));
              } // switch token.tag
            break; // case Node.RIGHT
            
          default:
            throw new ReteException("Node2::callNode",
                                    "Bad callType",
                                    String.valueOf(callType));
            
          } // switch callType
      }
    catch (ReteException re)
      {
        re.addContext("rule LHS: " + toString());
        throw re;
      }
    // the runTestsVary* routines pass messages on to the successors.

    return true;
  }

  /**
    Node2.callNode can call this to produce debug info.
    */

  void debugPrint(Token token, int callType) throws ReteException 
  {
    System.out.println("TEST " + toString() + ";calltype=" + callType
                       + ";tag=" + token.m_tag + ";class=" +
                       token.fact(0).get(RU.CLASS).stringValue());
  }

  /**
    Run all the tests on a given (left) token and every token in the
    right memory. For the true ones, assemble a composite token and
    pass it along to the successors.
    */

  boolean runTestsVaryRight(Token lt, TokenTree th) throws ReteException 
  {
    if (th == null)
      return false;

    TokenTree thl, thr;
    if ((thl = th.m_left) != null)
      if (runTestsVaryRight(lt, thl)) return true;
    if ((thr = th.m_right) != null)
      if (runTestsVaryRight(lt, thr)) return true;

    Token rt = th.m_token;

    while (rt != null)
      {
        if (runTests(lt, rt)) 
          { 
            Token nt = new Token(lt,rt.topFact());
            passAlong(nt);
            if (m_unique)
              return true;
          }
        rt = rt.m_next;
      }
    return false;
  }
    
  /**
    Run all the tests on a given (right) token and every token in the
    left memory. For the true ones, assemble a composite token and
    pass it along to the successors.
    */

  boolean runTestsVaryLeft(Token rt, TokenTree th) throws ReteException 
  {
    if (th == null)
      return false;

    if (th.m_left != null) runTestsVaryLeft(rt, th.m_left);
    if (th.m_right != null) runTestsVaryLeft(rt, th.m_right);

    Token lt = th.m_token;

    while (lt != null)
      {
        if (runTests(lt, rt)) 
          {
            // the new token has the *left* token's tag at birth...
            Token nt = new Token(lt,rt.topFact());
            nt.m_tag = rt.m_tag;
            passAlong(nt);
          }

        lt = lt.m_next;
      }
    return false;
  }


  /**
    Run all this node's tests on two tokens.  This routine assumes
    that the right token has only one fact in it, which is true at this
    time!
    */

  boolean runTests(Token lt, Token rt)
       throws ReteException 
  {
    int ntests = m_localTests.length;
    ValueVector rf = rt.topFact();

    EvalCache cache = m_cache;
    int markF = cache.markFuncall();
    int markV = cache.markValue();

    try
      {
        for (int i=0; i < ntests; i++) 
          {
            if (!m_localTests[i].doTest(lt, rf, cache, m_engine))
              return false;
          }        
        return true;
        
      }
    finally
      {
        cache.restoreFuncall(markF);
        cache.restoreValue(markV);
      }
  }

  /**
    Describe myself
    */
  
  public String toString() 
  {
    StringBuffer sb = new StringBuffer(256);
    sb.append("[Node2 ntests=");
    sb.append(m_tests.size());
    sb.append(" ");
    for (int i=0; i<m_tests.size(); i++)
      {
        sb.append(m_tests.elementAt(i).toString());
        sb.append(" ");
      }
    sb.append(";usecount = ");
    sb.append(m_usecount);
    sb.append(";unique = ");
    sb.append(m_unique);
    sb.append("]");
    return sb.toString();
  }


  public String displayMemory()
  {
    StringBuffer sb = new StringBuffer(256);
    sb.append("\n\nLeft Memory:\n\n");
    int count = showMemory(m_left, sb);
    sb.append ("\n");
    sb.append(count);
    sb.append(" entries");
    sb.append("\n\nRight Memory:\n\n");
    count = showMemory(m_right, sb);
    sb.append ("\n");
    sb.append(count);
    sb.append(" entries");
    return sb.toString();
  }

  private int showMemory(TokenTree th, StringBuffer sb)
  {
    if (th == null)
      return 0;
    int count = 0;

    count += showMemory(th.m_left, sb);

    if (th.m_token != null)
        {
          Token t = th.m_token;
          while (t != null)
            {
              ++count;
              sb.append(t);
              sb.append('\n');
              t = t.m_next;
            }
        }

    count += showMemory(th.m_right, sb);
    return count;
  }
  
}





